'''
Module contains spell and recipe for run adventure.
'''
from importlib import machinery
from os import listdir
from os.path import join, isdir

from .. import Spell
from ..recipe import Recipe, Ingredient
from .utils import apply
from ..types import LOCAL_SPELL
from .... import Habbit


class RunAdventure(Spell):
    name = 'run'
    description = 'Run travels of adventure'
    type = LOCAL_SPELL
    recipe = Recipe(
        Ingredient(synonyms=['-t', '--travels'], type=str, description='Name of travel.'),
        Ingredient(synonyms=['-l', '--lands'], type=str, description='Name of land.'),
        Ingredient(synonyms=['-s', '--sallies'], type=str, description='Name of sallies.'),
        Ingredient(synonyms=['-b', '--brotherhoods'], type=str,
                   description='Name of brotherhoods.'))

    def execute(self):
        self.travels = self.recipe.travels and self.recipe.travels.split(',')
        self.sallies = self.recipe.sallies and self.recipe.sallies.split(',') or set()
        self.brotherhoods = self.recipe.brotherhoods and self.recipe.brotherhoods.split(',') or set()
        apply(lambda land: self._run(land), self._get_lands())

    def _get_lands(self):
        return self.recipe.lands.split(',') if self.recipe.lands else [self.rucksack.TRAVELS_PATH, ]

    def _run(self, travel_path):
        """
        Run travels or land.
        """
        if isdir(travel_path):
            apply(lambda travel_file: self._run(join(travel_path, travel_file)), listdir(travel_path))
        elif travel_path.endswith('.py') and not travel_path.endswith('__init__.py'):
            self._run_travel(machinery.SourceFileLoader('travel', travel_path).load_module())

    def _run_travel(self, travel):
        if (not self.sallies) or ((self.sallies and 'SALLIES' in travel.__dict__)
                                        and self.sallies.issubset(travel.SALLIES)):
            apply(lambda habbit: self._run_habbit(habbit), self._get_habbits(travel))

    def _get_habbits(self, travel):
        return filter(lambda part: isinstance(part, type) and issubclass(part, Habbit)
                      and self._included_in_brotherhood(part), travel.__dict__.values())

    def _included_in_brotherhood(self, habbit):
        return True if 'brotherhoods' in habbit.__dict__ \
            and self.brotherhoods.issubset(habbit.brotherhoods) else False

    def _run_habbit(self, habbit):
        frodo = habbit(self.rucksack)
        frodo.run()



from os.path import dirname, abspath, join

PACKAGE_PATH = dirname(abspath(__file__))
PROJECT_TEMPLATE_PATH = join(PACKAGE_PATH, 'templates/adventure_template')
TRAVEL_TEMPLATE = join(PACKAGE_PATH, 'templates/travel_template.py')



"""
This module contains classes implementing the actions and structures
to get the results from the html page, or part of html page.
Classes describing the structure:
    - KeysRealtions; #A class describes the relationship keys or columns.
Classes describing the actions:
    - gtexts; #A class describes the action to get the text of the html item.
    - gattrs; #A class describes the action to get the value attribute of the page element.
    - gvalues; #A class describes the action to get the values for keyvalues pattern from html page.
"""
from functools import reduce
from operator import xor

from .actions import Action

# The string is passed to gvalues as KeysRelationships to produce a dictionary with original keys.
SOURCE_KEYS = 'source_keys'


"""
                                        Structure classes.
==================================================================================================
"""


class KeyRealtionships:
    """
    A class describes the relationship keys or columns.
    The class is the implementation of Hashed dictionary.
    A class used composite pattern.

    keys -> origin names of keys or columns;
    values -> target names of keys or columns;

    For the iteration used method items() of dictionary.
    """
    def __init__(self, relations):
        self.relations = relations

    def __hash__(self):
        hashed_keys = map(hash, self.relations.keys())
        hashed_values = map(hash, self.relations.values())
        return reduce(xor, hashed_keys) ^ reduce(xor, hashed_values)

    def __len__(self):
        return len(self.relations.keys())

    def __getitem__(self, key):
        return self.relations[key]

    def __repr__(self):
        return str(self.relations)

    def __str__(self):
        return str(self.relations)

    def __iter__(self):
        for key, value in self.relations.items():
            yield key, value
        return

    def keys(self):
        """
        Get keys as lists.
        """
        return self.relations.keys()


"""
                                        Action classes.
==================================================================================================
"""


class gtexts(Action):
    """
    A class describes the action to get the text of the html item.
    """
    def __init__(self, *args, **kwargs):
        """
        Initialization action.
        """
        super().__init__(*args, **kwargs)
        if 'separator' in kwargs:
            del self.kwargs['separator']
        self.separator = kwargs.get('separator') or ''

    def __call__(self, parent, key=None):
        """
        Run action.
        Input:
            parent -> tuple of url, BeautifulSoup Tag of parent;
            key -> Key of result dictionary;
        Output:
            result -> if key is not None then dict(key: result_list)
                      else result_list.
                      result_list is list with result texts from html elements;
        """
        parent_url, parent_tag = parent
        result_list = [tag.get_text(separator=self.separator).strip() for tag in
                       parent_tag.find_all(*self.args, **self.kwargs)
                       if self.filter_function(parent_url, tag)]
        result = result_list if len(result_list) > 1 else (result_list[0] if result_list else None)
        return {key: result} if key else result


class gattrs(Action):
    """
    A class describes the action to get the value attribute of the page element.
    """
    def __init__(self, *args, **kwargs):
        """
        Initialization action.
        Input:
            ....
            target_attribute -> Name of target attribute from html element;
        """
        super().__init__(*args, **kwargs)
        if 'target_attribute' in kwargs:
            del self.kwargs['target_attribute']
        self.target_attribute = kwargs.get('target_attribute') or 'alt'

    def __call__(self, parent, key=None):
        """
        Run action.
        Input:
            parent -> tuple of url, BeautifulSoup Tag of parent;
            key -> Key of result dictionary;
        Output:
            result -> if key is not None then dict(key: result_list)
                      else result_list.
                      result_list is list with value of attribute from html elements;
        """
        parent_url, parent_tag = parent
        result_list = [tag[self.target_attribute]
                       for tag in parent_tag.find_all(*self.args, **self.kwargs)
                       if self.filter_function(parent_url, tag)]
        result = result_list if len(result_list) > 1 else (result_list[0] if result_list else None)
        return {key: result} if key else result


class gvalues:
    """
    A class describes the action to get the values for keyvalues pattern from html page.
    """
    def __init__(self, keys_action, values_action):
        """
        Initialization action.
        keys_action -> Action for get keys;
        values_action -> Action for get values;
        """
        self.keys_action = keys_action
        self.values_action = values_action

    def __call__(self, parent, keys_relations):
        """
        Run action.
        Input:
            parent -> tuple of url, BeautifulSoup Tag of parent;
            keys_relations -> Relationships of key. Object of KeyRealtionships
                              or constants such as SOURCE_KEYS;
        Output:
            result -> Dictionary of result;
        """
        result_dict = None
        if keys_relations == SOURCE_KEYS:
            result_dict = {key: value for key, value in zip(self.keys_action(parent),
                                                            self.values_action(parent))}
        else:
            print('Get result from ', parent[0])
            result_dict = {keys_relations[key]: value
                           for key, value in zip(self.keys_action(parent),
                                                 self.values_action(parent))
                           if key in keys_relations.keys()}
        result_dict['url'] = parent[0]
        return result_dict



from re import match
from collections import namedtuple
from typing import List, NamedTuple, Any


class Ingredient(NamedTuple):
    synonyms: List[str]
    type: Any
    description: str
    default: Any = None
    value: Any = None


class Recipe:
    ingredients: List = []

    def __init__(self, *ingredients):
        if self._controll_ingredients(ingredients):
            self.ingredients = ingredients
        else:
            raise TypeError

    @staticmethod
    def _controll_ingredients(ingredients):
        ingredients_names = [synonym for ingredient in ingredients
                             for synonym in ingredient.synonyms]
        ingredients_names_count = map(
            lambda ingredient_name: ingredients_names.count(ingredient_name), ingredients_names)
        return False if list(filter(lambda ingredient_name_count: ingredient_name_count > 1,
                                    ingredients_names_count)) else True

    @staticmethod
    def _get_ingredients(ingredients):
        return {self._get_ingredient_name(Ingredient): ingredient for ingredient in ingredients}

    def __getattr__(self, attribute):
        return self._get_ingredient(attribute).value

    def __set__(self, attribute, value):
        self._get_ingredient(attribute).value = value

    def _get_ingredient(self, ingredient_name):
        target_ingreient = self._find_ingredient(ingredient_name)
        if not target_ingreient:
            raise AttributeError
        return target_ingreient

    def _find_ingredient(self, ingredient_name):
        target_ingredients = list(
            filter(lambda ingredient: self._is_ingredient(ingredient_name, ingredient.synonyms),
                   self.ingredients))
        return target_ingredients and target_ingredients[0]

    def _is_ingredient(self, ingredient_name, synonyms):
        return True if ingredient_name == self._get_filtered_synonym(
                        lambda synonym: not synonym.startswith('-'), synonyms) \
                    or ingredient_name == self._get_filtered_synonym(
                        lambda synonym: match(r'^--\w+', synonym), synonyms) \
                    or ingredient_name == self._get_filtered_synonym(
                        lambda synonym: match(r'^-\w+', synonym), synonyms) else False

    @staticmethod
    def _get_filtered_synonym(filter_function, synonyms):
        print(filter_function)
        print(synonyms)
        filtered_synonyms = list(filter(filter_function, synonyms))
        print(filtered_synonyms)
        return filtered_synonyms[0].lstrip('-') if filtered_synonyms else None

    def set_ingredients_values(self, values):
        for ingredient_name, ingredient_value in values.items():
            setattr(self, ingredient_name, ingredient_value)

    def __iter__(self):
        return self.ingredients.__iter__()



"""
The module contains constants for spell types.
**Spell types**
:param GLOBAL_SPELL: A spell to perform *OUTSIDE* the created adventure. (For example - creating an adventure)
:param LOCAL_SPELL: A spell to perform *INSIDE* the created adventure. (For example - go traveling)
"""
GLOBAL_SPELL = 'GLOBAL_SPELL'
LOCAL_SPELL = 'LOCAL_SPELL'



'''
                                       Spell and Spell Recipe.
==================================================================================================
**Spell** is function to perform a conversion with an adventure("path_to_mordor" project).
**Recipe** is class to describing the call of the spell.
**Ingredient** is class to describing the ingredients for spell.
This data is needed to build command-line parser data.
The Recipe is used as a parser of subparsers.
The Ingredient is used as an argument of a parser based on a SpellRecipe.
The functions of the spell are passed arguments:
    :param current_path: the path where the script was run;
    :param rucksack: module with adventure settings;
    :param ingredients: value of ingredients of spell;
    :type current_path: str
    :type rucksack: module or tuple
    :type ingredients: dict
'''
from os import getcwd
from collections import namedtuple

from .recipe import Recipe

from .types import GLOBAL_SPELL, LOCAL_SPELL


class Spell:
    name = 'spell'
    description = ''

    def __init__(self, rucksack):
        self.rucksack = rucksack
        self.type = (rucksack and LOCAL_SPELL) or GLOBAL_SPELL
        self.adventure_path = self._get_adventure_path()

        #recipe = Recipe([])

    def _get_adventure_path(self):
        return (self.rucksack and self.rucksack.ADVENTURE_PATH) or getcwd()

    def execute(self):
        pass



from argparse import ArgumentParser
from os import getcwd

from .spells.implementations import spells as _spells
from .spells.types import GLOBAL_SPELL
from ..ptm_settings import PACKAGE_PATH, PROJECT_TEMPLATE_PATH, TRAVEL_TEMPLATE


def execute_spell(spell_type=GLOBAL_SPELL, rucksack=None):
    spells = _get_spells(rucksack, spell_type)
    arguments = _parse_args(spells)
    if 'Spell' in arguments.__dict__:
        _run_spell(arguments, rucksack)
    else:
        _print_spells(spells)


def _run_spell(arguments, rucksack):
    spell = arguments.Spell(rucksack)
    spell.recipe.set_ingredients_values({key: value for key, value in arguments.__dict__.items()
                                         if key != 'Spell'})
    spell.execute()


def _print_spells(spells):
    print('Select one of spells:')
    for spell in spells:
        print('    * {}'.format(spell.name))


def _get_spells(rucksack, spell_type):
    return [Spell for Spell in _spells + (_get_extensions_spells(rucksack) or ())
            if Spell.type == spell_type]


def _parse_args(spells):
    parser = _create_argparse(spells)
    return parser.parse_args()


def _create_argparse(spells):
    parser = ArgumentParser(
        description='Character for adventure managment.(Adventure is "path_to_mordor" project)')
    subparsers = parser.add_subparsers()
    for Spell in spells:
        subparser = subparsers.add_parser(Spell.name)
        for ingredient in Spell.recipe:
            subparser.add_argument(
                *ingredient.synonyms, type=ingredient.type, help=ingredient.description)
        subparser.set_defaults(Spell=Spell)
    return parser


def _get_extensions_spells(rucksack):
    return (_get_extensions(rucksack) and _get_extensions(rucksack).__dict__.get('SPELLS')) or []


def _get_extensions(rucksack):
    return rucksack and rucksack.__dict__.get('EXTENSIONS')



'''
Spell and Recipe for start new adventure("path_to_mordor" project).
'''
from os import listdir, makedirs
from os.path import join, exists, isdir
from shutil import copy2, copytree

from .. import Spell
from ..recipe import Recipe, Ingredient
from ..types import GLOBAL_SPELL

from ....ptm_settings import PACKAGE_PATH, PROJECT_TEMPLATE_PATH


class StartAdventure(Spell):
    name = 'start_adventure'
    description = 'Create adventure("path_to_mordor" project).'
    type = GLOBAL_SPELL
    recipe = Recipe(
        Ingredient(synonyms=['-n', '--name', ], type=str, description='Name of adventure.'),
        Ingredient(synonyms=['-p', '--path'], type=str,
                   description='Path to adventure', default=''))

    def execute(self):
        self.adventure_path = self._get_adventure_path_with_recipe()
        if not exists(self.adventure_path) or not isdir(self.adventure_path):
            makedirs(self.adventure_path)

        # Create new adventure from adventure template.
        adventure_template_path = join(PACKAGE_PATH, PROJECT_TEMPLATE_PATH)
        self._create_adventure(adventure_template_path)

    def _get_adventure_path_with_recipe(self):
        adventure_path = self._join_adventure_path(self.adventure_path, self.recipe.path)\
            if self.recipe.path else self.adventure_path
        return join(adventure_path, self.recipe.name) if self.recipe.name else adventure_path

    @staticmethod
    def _join_adventure_path(current_path: str, path: str):
        return path if path.startswith('/') else join(current_path, path)

    def _create_adventure(self, adventure_template_path: str):
        for template_path in listdir(adventure_template_path):
            self._copy_template(join(adventure_template_path, template_path),
                                join(self.adventure_path, template_path))

    @staticmethod
    def _copy_template(absolute_template_path: str, absolute_adventure_path: str):
        if not exists(absolute_adventure_path):
            if isdir(absolute_template_path):
                copytree(absolute_template_path, absolute_adventure_path)
            else:
                copy2(absolute_template_path, absolute_adventure_path)



"""
Module contains class to get results.
To fech results used BeautifulSoup.
"""
from functools import reduce
from operator import xor

class ResultAction: #pylint: disable-msg=R0903
    """
    Base class for describing action of get result.
    """
    def __init__(self, *args, **kwargs):
        """
        Initial.
            HTML block description
            ----------------------
            Use arguments and named arguments to find_all() method from BeautifulSoup.
            args -> Argumnets for find_all(), like as html block name - 'a', 'div' and etc;
            kwargs -> Named arguments for find_all(), like as "attrs={'class': 'some_css_class'}";
        """
        self.args = args
        self.kwargs = kwargs

class gtexts(ResultAction): #pylint: disable-msg=R0903,C0103
    """
    Class for decribing action of get text from BeautifulSoup Tag.
    """
    def __init__(self, *args, **kwargs):
        """
        Initial of super initial.
        """
        super().__init__(*args, **kwargs)

    def __call__(self, parent_tag, key):
        """
        Run action.
        Input:
            parent_tag -> BeautifulSoup Tag;
        Output:
            list_of_text -> List of texts from BeautifulSoup Tags;
        """
        return {key: [tag.get_text().strip() for tag in parent_tag.find_all(*self.args, **self.kwargs)]}

class gattrs(ResultAction): #pylint: disable-msg=R0903,C0103
    """
    Class for describing action of get attributes from BeautifulSoup Tag.
    """
    def __init__(self, *args, target_attribute='alt', **kwargs):
        """
        Initial of super initial.
        Added target atribute.
        Input:
            target_attribute -> Target attribute;
        """
        super().__init__(*args, **kwargs)
        self.target_attribute = target_attribute

    def __call__(self, parent_tag, key=None):
        """
        Run action.
        Input:
            parent_tag -> Parent BeautifulSoup Tag;
        Output:
            list_of_attributes -> List of value of atributes;
        """
        return {key: [tag[self.target_attribute]
                for tag in parent_tag.find_all(*self.args, **self.kwargs)]}

class KeysRealtions: #pylint: disable-msg=R0903
    """
    Class for describing relations of keys.
    """
    def __init__(self, relations):
        self.relations = relations

    def __hash__(self):
        return reduce(xor, self.relations.values().sort())

class KeyValue: #pylint: disable-msg=R0903
    """
    Class to describig key-value relations.
    """
    def __init__(self, keys_action, values_action):
        self.keys_action = keys_action
        self.values_action = values_action

    def __call__(self, parent_tag, keys_relations):
        relations = keys_relations.relations
        return {relations[key]: value for key, value in zip(self.keys_action(parent_tag, 'key')['key'], self.values_action(parent_tag, 'key')['key']) if key in relations.keys()}



'''
Module contains spell and recipe for create travel. Travel is python module
containing rules for scrapping.
'''
from os.path import join, isfile
from shutil import copy2

from .. import Spell
from ..recipe import Recipe, Ingredient
from ..types import LOCAL_SPELL

from ....ptm_settings import TRAVEL_TEMPLATE


class CreateTravel(Spell):
    name = 'create_travel'
    description = '''Create travel. Travel is a python module
                          containing scrapping rules.'''
    type = LOCAL_SPELL

    recipe = Recipe(Ingredient(synonyms=['name', ], type=str, description='Name of travel.'),
                    Ingredient(synonyms=['-l', '--land'], type=str,
                               description='Land of travel.', default=''))

    def execute(self):
        land_path = self.recipe.land or ''
        travel_path = join(join(self.rucksack.TRAVELS_PATH, land_path),
                           '{}.py'.format(self.recipe.name))
        if not isfile(travel_path):
            copy2(TRAVEL_TEMPLATE, travel_path)
        else:
            print('Package {} is exist!'.format(travel_path))



#pylint: disable-msg=C0103
"""
Module contains classes to scraping.
"""
from time import sleep
from re import match
from concurrent import futures
import urllib.request

from bs4 import BeautifulSoup


__version__ = '0.0.1a4'


"""
                                            FrHabbitodo.
==================================================================================================
"""


class Habbit:
    start_page_url = ''
    path = {}

    """
    Class performs scraping.
    """
    def __init__(self, rucksack):
        """
        Initial.
        Input:
            start_page_url -> Url of start page to scraping;
            path -> Path of scraping;
            ruksack -> Rucksack of Habbit; (Scraping settings);
        """
        self.rucksack = rucksack
        self.max_concurrent_steps = rucksack.MAX_CONCURRENT_STEPS
        self.wait_time = rucksack.WAIT_TIME
        self.render_wait_time = rucksack.RENDER_WAIT_TIME
        self.current_concurrent_steps = 0
        self.browser = self._get_browser(rucksack)

        self._set_resource_from_url(self.start_page_url)
        self._start_steps = self._create_steps()

    @staticmethod
    def _get_browser(rucksack):
        if 'BROWSER' in rucksack.__dict__.keys() or not rucksack.BROWSER:
            browser = BrowserSubstitute
        else:
            browser = rucksack.BROWSER
        return browser

    def _set_resource_from_url(self, url):
        try:
            self.resource = match(r'^http[s]?://.*\.\w{1,3}', url).group(0)
        except ValueError:
            print(ValueError)

    """
                                            Get region.
    ==============================================================================================
    Region is a text of html page.
    """
    def get_region(self, url):
        """
        Get page of text from url.
        Input:
            url -> Url address of page;
        Output:
            page_text -> Source text of html page;
        """
        self._grab_region()
        browser = self.browser()
        browser.get(url)
        begin_page_text = browser.page_source
        self._release_region()
        if self.render_wait_time and begin_page_text == browser.page_source:
            sleep(self.render_wait_time)
        page_text = browser.page_source
        browser.close()
        return page_text

    def _grab_region(self):
        """
        Wait one's turn.
        Increase counter self.current_concurrent_steps with grab.
        """
        while self.current_concurrent_steps >= self.max_concurrent_steps:
            sleep(self.wait_time)
        self.current_concurrent_steps += 1

    def _release_region(self):
        """
        Release region.
        """
        self.current_concurrent_steps -= 1

    """
                                            Create steps.
    ==============================================================================================
    """
    def _create_steps(self):
        """
        Create lists of steps.
        """
        steps = []
        for action, next_step in self.path.items():
            steps.append(Step(action,
                              self._create_steps(next_step) if isinstance(next_step, dict)
                              else [Step(next_step, [])]))
        return steps

    """
                                            Run adventure.
    ==============================================================================================
    """

    def run(self):
        """
        Run scraping.
        """
        for step in self._start_steps:
            self._make_step(step,
                            (self.start_page_url,
                             BeautifulSoup(self.get_region(self.start_page_url), 'html.parser')))

    def _make_step(self, step, parent):
        """
        Make step.
        """
        for result in step.current_action(self, parent):
            for next_step in step.next_steps:
                self._make_step(next_step, result)


class Step:
    """
    Class for describing step of scraping,
    """
    def __init__(self, current_action, next_steps=None):
        self.current_action = current_action
        self.next_steps = next_steps or []

    def __str__(self):
        return '(Current action: {}, Next step: {})'.format(self.current_action, self.next_steps)

    def __repr__(self):
        return '(Current action: {}, Next step: {})'.format(self.current_action, self.next_steps)


class BrowserSubstitute:
    def __init__(self):
        self.page_source = ""

    def get(self, url):
        with urllib.request.urlopen(url) as request:
            return request.read()

    def close(self):
        pass



"""
Module to manage and run adventure.
"""
import rucksack
from ptm.adventure_managment import execute_spell
from ptm.adventure_managment.spells.types import LOCAL_SPELL


def execute():
    execute_spell(LOCAL_SPELL, rucksack)


if __name__ == '__main__':
    execute()



"""
This module contains classes implementing the actions to scraping.
Classes describing the actions:
    - gpages; #A class describes the action to get pages from url from html elements.
    - gpagins; #A class describes the action to get pagination pages.
    - gblocks; #A class describes the action to get html from html page.
    - gresults; #A class describes the action to get results from ftml pages.
"""
from . import Action, ThreadAction


"""
                                    Actions with load of pages.
==================================================================================================
"""


class gpages(ThreadAction):
    """
    A class describes the action to get pages from url from html elements.
    """
    def __init__(self, *args, **kwargs):
        """
        Initialization action.
        """
        super().__init__(*args, **kwargs)

    def __call__(self, frodo, parent):
        """
        Run action.
        Input:
            frodo -> Object of Frodo trgerring action;
            parent -> parent_url, parent_tag(BeautifulSoup Tag);
        Output:
            yeild url, tag_of_page -> url of page, BeautifulSoup Tag of page;
        """
        _, parent_tag = parent
        target_tags = parent_tag.find_all(*self.args, **self.kwargs) if self.args else [parent_tag]
        urls = self._get_urls(frodo.resource, target_tags)
        yield from self.get_pages(frodo, urls)

    @staticmethod
    def _get_urls(resource, target_tags):
        """
        Get urls from target_tags(BeautifulSoup Tags).
        Input:
            resource -> Url of resource;
            target_tags -> List of BeautifulSoup Tags;
        """
        return ['{resource}{href}'.format(resource=resource, href=tag['href'])
                for target_tag in target_tags for tag in target_tag.find_all('a')]


class gpagins(ThreadAction):
    """
    Class decribing action of getting pagination pages.
    """
    def __init__(self, pagin_template, filter_function=None,
                 start_page_number=None, finish_page_number=None):
        """
        Initialization action.
        Initialization pagination fields.
        Input:
            super...
            pagin_template -> Template of pagination, like as "?page=";
            start_page_number -> Number of start page;
            finish_page_number -> Number of finish page;
        """
        super().__init__(filter_function=filter_function)
        self.pagin_template = pagin_template
        self.start_page_number = start_page_number if start_page_number else 1
        self.finish_page_number = finish_page_number if finish_page_number else 1000000

    def __call__(self, frodo, parent):
        """
        Run action.
        Input:
            frodo -> Object of Frodo trgerring action;
            parent -> parent_url, parent_tag(BeautifulSoup Tag);
        Output:
            yeild url, tag_of_page -> url of page, BeautifulSoup Tag of page;
        """
        parent_url, _ = parent
        urls = self._get_urls(parent_url)
        yield from self.get_pages(frodo, urls)

    def _get_urls(self, parent_url):
        """
        Get pagination urls.
        Input:
            parent_url -> Url to pagination;
        """
        return ['{parent_url}/{pagin_template}{page_number}'.format(
                        parent_url=parent_url,
                        pagin_template=self.pagin_template,
                        page_number=page_number)
                for page_number in range(self.start_page_number, self.finish_page_number + 1)]

    def __hash__(self):
        start_page_hash = self._hash_or_empty(self.start_page_number)
        finish_page_hash = self._hash_or_empty(self.finish_page_number)
        filter_function_hash = self._filter_function_hash()
        return start_page_hash ^ finish_page_hash ^ filter_function_hash ^ hash(self.pagin_template)


"""
                                    Actions with finding html blocks.
==================================================================================================
"""


class gblocks(Action):
    """
    Class for find BeautifulSoup Tag in parent BeautifulSoup Tag.
    Class is the interface to BeautifulSoup "find_all".
    """
    def __init__(self, *args, **kwargs):
        """
        Initial of super initial of Action.
        """
        super().__init__(*args, **kwargs)

    def __call__(self, frodo, parent):
        """
        Run action.
        Input:
            frodo -> Object of Frodo trgerring action;
            parent -> parent_url, parent_tag(BeautifulSoup Tag);
        Output:
            yeild url, tag_of_page -> Url of page, BeautifulSoup Tag of page;
        """
        parent_url, parent_tag = parent
        for tag in parent_tag.find_all(*self.args, **self.kwargs):
            if self.filter_functions(tag):
                yield (parent_url, tag)


"""
                                        Get results.
==================================================================================================
"""


class gresults:
    """
    Class to get result from parent BeautifulSoup Tag.
    """
    def __init__(self, result_map=None):
        """
        Inital.
        Input:
            result_function -> Function to procces results;
                Input:
                    result_set -> dictionary with results;
            result_map -> Dictionary with result fields(Keys) and result action;
        """
        self.result_map = result_map

    def __call__(self, frodo, parent):
        """
        Run action.
        Input:
            parent -> Tuple of url, parent BeautifulSoup Tag;
        """
        url, _ = parent
        result_set = {}
        for result_name, result_action in self.result_map.items():
            result_set.update(result_action(parent, result_name))
        result_set['url'] = url
        frodo.pick_up(result_set)
        yield

    def __hash__(self):
        return hash('results')



"""
Module contains settings of adventure.
"""
from os.path import dirname, abspath, join

import extensions

# Project path.
ADVENTURE_PATH = dirname(abspath(__file__))

# Travels path,
TRAVELS_PATH = join(ADVENTURE_PATH, 'travels')

# Treasy path.
TREASY_PATH = join(ADVENTURE_PATH, 'treasy')

# The maximum number of running browsers.
MAX_CONCURRENT_STEPS = 4

# Waiting time between grab for browser for geting page text.
WAIT_TIME = 0.5

# Browser(with selenium webdriver) to get html from urls.
# Like as PhontomJS, Chrome and etc...
# If None then used urllib.request.
BROWSER = None

# Time to render js(sec.).
RENDER_WAIT_TIME = 2

EXTENSIONS = extensions



'''
Module contains secondary functions for spells.
'''
from os import makedirs
from os.path import exists, join, isdir


def make_package(package_path: str):
    '''
    Create python package with package_path path.
        :param package_path: path of new package
        :type package_path: string
    '''
    if not exists(package_path) or not isdir(package_path):
        makedirs(package_path)
        make_module(join(package_path, '__init__.py'))


def make_module(module_path: str):
    '''
    Create python module.
        :param module_path: path of new module
        :type module_path: string
    '''
    if not exists(module_path):
        with open(module_path, 'w+'):
            pass


def apply(function, iterator):
    '''
    Apply function for all values from iterator.
    '''
    for value in iterator:
        function(value)



'''
Module contains spell and recipe for create land. Land is python package containing travels(Python
modules containing scraping rules).
'''
from os.path import join, isdir

from .. import Spell
from ..recipe import Recipe, Ingredient
from ..types import LOCAL_SPELL
from .utils import make_package


class CreateLand(Spell):
    name = 'create_land'
    description = '''Create land. Land is group of travels.
                          Travel is python module. Land is a python
                          package containing modules(travels)
                          or other python packages(lands).'''
    type = LOCAL_SPELL
    recipe = Recipe(Ingredient(synonyms=['name', ], type=str, description='Name of land.'))

    def execute(self):
        land_path = join(self.rucksack.TRAVELS_PATH, self.recipe.name)
        if not isdir(land_path):
            make_package(land_path)
        else:
            print('Path {} is exist.'.format(land_path))



"""
The module contains the rules of scraping.
"""
from ptm import Habbit


SALLIES = ()


class Frodo(Habbit):
    brotherhoods = ()
    start_page_url = ''
    path = {}

    def pcik_up(self, result_set):
        pass



from .start_adventure import StartAdventure
from .create_land import CreateLand
from .create_travel import CreateTravel
from .run_adventure import RunAdventure

spells = (
    StartAdventure,
    CreateLand,
    CreateTravel,
    RunAdventure
)



"""
Module contains classes descripbing actions of scraping.
For parsing used BeautifulSoup.
BeautifulSoup documetnation: https://www.crummy.com/software/BeautifulSoup/bs4/doc/.
"""
from concurrent.futures import ThreadPoolExecutor, as_completed
from functools import reduce
from operator import xor

from bs4 import BeautifulSoup


"""
                                        Base Action classes.
==================================================================================================
"""


class Action:
    """
    The base class for scrapping actions.
    """
    def __init__(self, *args, **kwargs):
        """
        Input:
            filter_function -> Function to filtering result of action;
                               Input:
                                    url, bs_tag;

            HTML block description
            ----------------------
            Use arguments and named arguments to find_all() method from BeautifulSoup.
            args -> Argumnets for find_all(), like as html block name - 'a', 'div' and etc;
            kwargs -> Named arguments for find_all(), like as "attrs={'class': 'some_css_class'}";
        """
        self.args = args
        self.kwargs = kwargs
        self.filter_function = kwargs.get('filter_function') or (lambda url, bs_tag: True)

    """
    ____________________________________________hash__________________________________________
    """

    def __hash__(self):
        kwargs_hash = self._get_kwargs_hash(self.kwargs)
        args_hash = reduce(xor, map(hash, self.args)) if self.args else hash('')
        filter_function_hash = self._filter_function_hash()
        return kwargs_hash ^ args_hash ^ filter_function_hash

    def _get_kwargs_hash(self, kwargs):
        kwargs_hash = hash('')
        for key, value in kwargs.items():
            kwargs_hash ^= self._get_kwargs_hash(value) if isinstance(value, dict)\
                                                   else hash(key) ^ hash(value)
        return kwargs_hash

    @staticmethod
    def _hash_or_empty(value):
        return hash(value) if value else hash('')

    def _filter_function_hash(self):
        return hash(self.filter_function.__name__) if self.filter_function else hash('')


class ThreadAction(Action):
    """
    The base class for the actions of working with multiple threads.
    Thread used to load pages with Frodo.
    """
    def __init__(self, *args, **kwargs):
        """
        Init is similar to the Action init. Added max_workers variable to limit threads count.
        """
        super().__init__(*args, **kwargs)
        self.max_workers = 20

    def get_pages(self, frodo, urls):
        """
        Get BeautifulSoup Tag of pages from urls.
        Input:
            frodo -> Frodo object for load page_text;
            urls -> List of urls;
        Output:
            yeild url, tag_of_page -> url of page, BeautifulSoup Tag of page;
        """
        check_page_repeat = CheckPageRepeat()
        with ThreadPoolExecutor(max_workers=self.max_workers) as executor:
            future_set = {executor.submit(frodo.get_region, url): url for url in urls}
            for future in as_completed(future_set):
                page_text = future.result()
                page_tag = BeautifulSoup(page_text, 'html.parser')
                if not check_page_repeat(page_text)\
                   and self.filter_function(future_set[future], page_tag):
                    yield (future_set[future], page_tag)
                elif check_page_repeat.check_break:
                    break
        # for url in urls:
        #     yield url, BeautifulSoup(frodo.get_region(url), 'html.parser')


class CheckPageRepeat:
    """
    Class to check of page repeat.
    """
    def __init__(self, max_page_repeat=3):
        """
        Initial.
        Input:
            max_page_repeat -> Maximum repeat of page.
                               If repeat is maximum then check_break is True;
        """
        self.__last_page_text = None
        self.__repeat_page = 0
        self.__check_break = False
        self.max_page_repeat = max_page_repeat

    @property
    def check_break(self):
        """
        Break if repeat.
        """
        return self.__check_break

    def __call__(self, page_text):
        """
        Check repeat page.
        Input:
            page_text -> Text of html page;
        Output:
            True, if repeat, False if not;
        """
        repeat = False
        if page_text == self.__last_page_text:
            self.__repeat_page += 1
            repeat = True
        else:
            self.__repeat_page = 0
            self.__last_page_text = page_text
        if self.__repeat_page > self.__repeat_page:
            self.__check_break = True
        return repeat



from setuptools import setup, find_packages
from os.path import join, dirname

from ptm import __version__ as version

setup(
    name='ptm',
    version=version,
    packages=find_packages(),
    author='acrius',
    author_email='acrius@mail.ru',
    url='https://github.com/acrius/path_to_mordor',
    description='',
    license='MIT',
    keywords='',
    long_description=open(join(dirname(__file__), 'README.md')).read(),
    include_package_data=True,
    package_data={'templates': ['templates/*', ]},
    entry_points={
        'console_scripts': ['galadriel = ptm.adventure_managment:execute_spell']
    },
    include_packages_data=True,
    install_requires=[
        'beautifulsoup4'
    ]
)
